using System.Net.WebSockets;
using Nostr.Client.Identifiers;
using Nostr.Client.Messages;
using Nostr.Client.Utils;
using NostrRelay;
using NostrServices.Services;
using ProtoBuf;
using StackExchange.Redis;

namespace NostrServices;

public static class Extensions
{
    public static string? GetFirstTagValue(this CompactEvent ev, string key)
    {
        return ev.Tags.FirstOrDefault(a => a.Key == key)?.Values[0];
    }

    public static async Task<bool> SetAsync<T>(this IDatabase db, RedisKey key, T val, TimeSpan? expire = null)
    {
        using var ms = new MemoryStream();
        Serializer.Serialize(ms, val);
        return await db.StringSetAsync(key, ms.ToArray(), expire);
    }

    public static async Task<T?> GetAsync<T>(this IDatabase db, RedisKey key) where T : class
    {
        var data = await db.StringGetAsync(key);
        if (data is {HasValue: true, IsNullOrEmpty: false})
        {
            return Serializer.Deserialize<T>(((byte[])data!).AsSpan());
        }

        return default;
    }

    public static NostrIdentifier ToIdentifier(this NostrEvent ev)
    {
        if ((long)ev.Kind is >= 30_000 and < 40_000)
        {
            return new NostrAddressIdentifier(ev.Tags!.FindFirstTagValue("d")!, ev.Pubkey!, [], ev.Kind);
        }

        return new NostrEventIdentifier(ev.Id!, ev.Pubkey, [], ev.Kind);
    }

    public static NostrIdentifier ToIdentifier(this CompactEvent ev)
    {
        if (ev.Kind is >= 30_000 and < 40_000)
        {
            return new NostrAddressIdentifier(ev.GetFirstTagValue("d")!, ev.PubKey.ToHex(), [], (NostrKind)ev.Kind);
        }

        return new NostrEventIdentifier(ev.Id.ToHex(), ev.PubKey.ToHex(), [], (NostrKind)ev.Kind);
    }

    public static NostrIdentifier ToIdentifier(this CompactProfile px)
    {
        return new NostrProfileIdentifier(px.PubKey.ToHex(), []);
    }

    public static void MapNostrRelay<THandler>(this IEndpointRouteBuilder app, string path) where THandler : INostrRelay
    {
        app.MapGet(path, async ctx =>
        {
            if (ctx.WebSockets.IsWebSocketRequest)
            {
                var logger = app.ServiceProvider.GetRequiredService<ILogger<NostrRelay<THandler>>>();
                var handler = app.ServiceProvider.GetRequiredService<THandler>();
                try
                {
                    var ws = await ctx.WebSockets.AcceptWebSocketAsync();
                    var wsCtx = new NostrClientContext(ws, ctx.Connection.RemoteIpAddress!,
                        ctx.Request.Headers.UserAgent.FirstOrDefault() ?? string.Empty);

                    var nostrRelay = new NostrRelay<THandler>(handler, wsCtx, ctx.RequestAborted, logger);

                    await nostrRelay.Read();

                    await ws.CloseAsync(WebSocketCloseStatus.NormalClosure, null, ctx.RequestAborted);
                }
                catch (Exception ex)
                {
                    logger.LogDebug(ex.Message);
                }
            }
        });
    }
}
